@@ -1,5 +1,5 @@
 /* gifread.c - Functions to read GIFs.
-   Copyright (C) 1997 Eddie Kohler, eddietwo@lcs.mit.edu
+   Copyright (C) 1997-9 Eddie Kohler, eddietwo@lcs.mit.edu
    This file is part of the GIF library.
 
    The GIF library is free software*. It is distributed under the GNU Public
@@ -11,7 +11,12 @@
    holder, allows the compression algorithm to be used without a license in
    software distributed at no cost to the user. */
 
-/* #define GIF_DEBUGGING */
+#ifdef HAVE_CONFIG_H
+# include <config.h>
+#elif !defined(__cplusplus)
+/* Assume we don't have inline by default */
+# define inline
+#endif
 #include "gif.h"
 #include <stdarg.h>
 #include <assert.h>
@@ -68,10 +73,7 @@ Gif_Debug(char *format, ...)
 #define gifgetblock(ptr, size, grr) ((*grr->block_getter)(ptr, size, grr))
 #define gifeof(grr)	((*grr->eofer)(grr))
 
-#ifdef __GNUC__
-__inline__
-#endif
-static u_int16_t
+static inline u_int16_t
 gifgetunsigned(Gif_Reader *grr)
 {
   byte one = gifgetbyte(grr);
@@ -172,23 +174,23 @@ read_image_block(Gif_Reader *grr, byte *buffer, int *bit_pos_store,
 {
   int bit_position = *bit_pos_store;
   int bit_length = *bit_len_store;
-  int i;
+  byte block_len;
   
   while (bit_position + bits_needed > bit_length) {
     /* Read in the next data block. */
     if (bit_position >= 8) {
       /* Need to shift down the upper, unused part of `buffer' */
-      i = bit_position / 8;
+      int i = bit_position / 8;
       buffer[0] = buffer[i];
       buffer[1] = buffer[i+1];
       bit_position -= i * 8;
       bit_length -= i * 8;
     }
-    i = gifgetbyte(grr);
-    GIF_DEBUG(("image_block(%d)", i));
-    if (i == 0) return 0;
-    gifgetblock(buffer + bit_length / 8, i, grr);
-    bit_length += i * 8;
+    block_len = gifgetbyte(grr);
+    GIF_DEBUG(("image_block(%d)", block_len));
+    if (block_len == 0) return 0;
+    gifgetblock(buffer + bit_length / 8, block_len, grr);
+    bit_length += block_len * 8;
   }
   
   *bit_pos_store = bit_position;
@@ -232,7 +234,7 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
   codemask = 0;
   for (code = 0; code < clear_code; code++) {
     gfc->prefix[code] = 49428;
-    gfc->suffix[code] = code;
+    gfc->suffix[code] = (byte)code;
     gfc->length[code] = 1;
   }
   eoi_code = clear_code + 1;
@@ -271,7 +273,7 @@ read_image_data(Gif_Context *gfc, Gif_Reader *grr)
     accum = buffer[i] + (buffer[i+1] << 8);
     if (bits_needed >= 8)
       accum |= (buffer[i+2]) << 16;
-    code = (accum >> (bit_position % 8)) & codemask;
+    code = (Gif_Code)((accum >> (bit_position % 8)) & codemask);
     bit_position += bits_needed;
     
     /* CHECK FOR SPECIAL OR BAD CODES: clear_code, eoi_code, or a code that is
@@ -399,9 +401,9 @@ read_logical_screen_descriptor(Gif_Stream *gfs, Gif_Reader *grr)
 static int
 read_compressed_image(Gif_Image *gfi, Gif_Reader *grr, int read_flags)
 {
-  if (grr->is_record && 0) {
+  if (grr->is_record) {
     byte *first = grr->v;
-    int pos;
+    u_int32_t pos;
     
     /* scan over image */
     pos = 1;			/* skip min code size */
@@ -639,8 +641,9 @@ read_unknown_extension(Gif_Stream *gfs, int kind, char *app_name, int position,
     data_len += block_len;
     block_len = gifgetbyte(grr);
   }
-  
-  gfex = Gif_NewExtension(kind, app_name);
+
+  if (data)
+    gfex = Gif_NewExtension(kind, app_name);
   if (gfex) {
     gfex->data = data;
     gfex->length = data_len;
@@ -817,7 +820,7 @@ read_gif(Gif_Reader *grr, int read_flags)
   /* Move comments after last image into stream. */
   gfs->comment = gfi->comment;
   gfi->comment = 0;
-
+  
   Gif_DeleteImage(gfi);
   Gif_DeleteArray(last_name);
   Gif_DeleteArray(gfc.prefix);
@@ -845,7 +848,6 @@ Gif_FullReadFile(FILE *f, int read_flags)
   return read_gif(&grr, read_flags);
 }
 
-
 Gif_Stream *
 Gif_FullReadRecord(Gif_Record *gifrec, int read_flags)
 {
