@@ -1,5 +1,5 @@
 /* gifsicle.c - gifsicle's main loop.
-   Copyright (C) 1997-8 Eddie Kohler, eddietwo@lcs.mit.edu
+   Copyright (C) 1997-9 Eddie Kohler, eddietwo@lcs.mit.edu
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
@@ -7,6 +7,7 @@
    as this notice is kept intact and this source code is made available. There
    is no warranty, express or implied. */
 
+#include <config.h>
 #include "gifsicle.h"
 #include <string.h>
 #include <stdio.h>
@@ -15,6 +16,13 @@
 #include <assert.h>
 #include <errno.h>
 
+/* Need _setmode under MS-DOS, to set stdin/stdout to binary mode */
+/* Need _fsetmode under OS/2 for the same reason */
+#if defined(_MSDOS) || defined(_WIN32) || defined(__EMX__)
+# include <fcntl.h>
+# include <io.h>
+#endif
+
 
 Gt_Frame def_frame;
 
@@ -99,13 +107,14 @@ static int any_output_successful = 0;
 #define CH_COLORMAP		4
 #define CH_DITHER		5
 #define CH_USE_COLORMAP		6
-#define CH_BACKGROUND		7
-#define CH_COLOR_TRANSFORM	8
-#define CH_RESIZE		9
+#define CH_COLORMAP_METHOD	7
+#define CH_BACKGROUND		8
+#define CH_COLOR_TRANSFORM	9
+#define CH_RESIZE		10
 static const char *output_option_types[] = {
   "loopcount", "logical screen", "optimization", "output file",
-  "colormap size", "dither", "colormap", "background",
-  "color transformation", "resize"
+  "colormap size", "dither", "colormap", "colormap method",
+  "background", "color transformation", "resize"
 };
 
 
@@ -162,6 +171,7 @@ static const char *output_option_types[] = {
 #define EXTENSION_OPT		352
 #define COLOR_TRANSFORM_OPT	353
 #define RESIZE_OPT		354
+#define SCALE_OPT		355
 
 #define LOOP_TYPE		(Clp_MaxDefaultType + 1)
 #define DISPOSAL_TYPE		(Clp_MaxDefaultType + 2)
@@ -172,6 +182,7 @@ static const char *output_option_types[] = {
 #define RECTANGLE_TYPE		(Clp_MaxDefaultType + 7)
 #define TWO_COLORS_TYPE		(Clp_MaxDefaultType + 8)
 #define COLORMAP_ALG_TYPE	(Clp_MaxDefaultType + 9)
+#define SCALE_FACTOR_TYPE	(Clp_MaxDefaultType + 10)
 
 Clp_Option options[] = {
   
@@ -185,13 +196,12 @@ Clp_Option options[] = {
   { "change-color", 0, CHANGE_COLOR_OPT, TWO_COLORS_TYPE, Clp_Negate },
   { "cinfo", 0, COLOR_INFO_OPT, 0, Clp_Negate },
   { "clip", 0, CROP_OPT, RECTANGLE_TYPE, Clp_Negate },
-  { "colors", 'k', COLORMAP_OPT, Clp_ArgInt,
-    Clp_Negate | Clp_LongMinMatch, 3 }, /****/
+  { "colors", 'k', COLORMAP_OPT, Clp_ArgInt, Clp_Negate },
   { "color-method", 0, COLORMAP_ALGORITHM_OPT, COLORMAP_ALG_TYPE, 0 },
   { "color-info", 0, COLOR_INFO_OPT, 0, Clp_Negate },
-  { "comment", 'c', COMMENT_OPT, Clp_ArgString, Clp_Negate },
+  { "comment", 'c', COMMENT_OPT, Clp_ArgString, 0 },
+  { "no-comments", 'c', NO_COMMENTS_OPT, 0, Clp_OnlyNegated },
   { "crop", 0, CROP_OPT, RECTANGLE_TYPE, Clp_Negate },
-  { "no-comments", 0, NO_COMMENTS_OPT, 0, Clp_LongMinMatch, 6 }, /****/
   
   { "delay", 'd', 'd', Clp_ArgInt, Clp_Negate },
   { "delete", 0, DELETE_OPT, 0, 0 },
@@ -199,15 +209,15 @@ Clp_Option options[] = {
   { "dither", 'f', DITHER_OPT, 0, Clp_Negate },
   { "done", 0, ALTER_DONE_OPT, 0, 0 },
   
-  { "explode", 'e', 'e', 0, Clp_LongMinMatch, 3 }, /****/
+  { "explode", 'e', 'e', 0, 0 },
   { "explode-by-name", 'E', 'E', 0, 0 },
-  { "extension", 0, EXTENSION_OPT, Clp_ArgString, Clp_LongMinMatch, 3 }, /***/
+  { "extension", 0, EXTENSION_OPT, Clp_ArgString, 0 },
+  { "no-extensions", 0, NO_EXTENSIONS_OPT, 0, 0 },
   { "extension-info", 0, EXTENSION_INFO_OPT, 0, Clp_Negate },
-  { "no-extensions", 0, NO_EXTENSIONS_OPT, 0, Clp_LongMinMatch, 5 }, /****/
   
   { "flip-horizontal", 0, FLIP_HORIZ_OPT, 0, Clp_Negate },
   { "flip-vertical", 0, FLIP_VERT_OPT, 0, Clp_Negate },
-  { "no-flip", 0, NO_FLIP_OPT, 0, Clp_LongMinMatch, 4 }, /****/
+  { "no-flip", 0, NO_FLIP_OPT, 0, 0 },
   
   { "help", 'h', HELP_OPT, 0, 0 },
   
@@ -221,9 +231,9 @@ Clp_Option options[] = {
   { "merge", 'm', 'm', 0, 0 },
   { "method", 0, COLORMAP_ALGORITHM_OPT, COLORMAP_ALG_TYPE, 0 },
   
-  { "name", 'n', NAME_OPT, Clp_ArgString, Clp_Negate },
+  { "name", 'n', NAME_OPT, Clp_ArgString, 0 },
+  { "no-names", 'n', NO_NAME_OPT, 0, Clp_OnlyNegated },
   { "netscape-workaround", 0, NETSCAPE_WORK_OPT, 0, Clp_Negate },
-  { "no-names", 0, NO_NAME_OPT, 0, Clp_LongMinMatch, 5 }, /****/
   
   { "optimize", 'O', OPTIMIZE_OPT, Clp_ArgInt, Clp_Negate | Clp_Optional },
   { "output", 'o', OUTPUT_OPT, Clp_ArgStringNotOption, 0 },
@@ -235,8 +245,9 @@ Clp_Option options[] = {
   { "rotate-90", 0, ROTATE_90_OPT, 0, 0 },
   { "rotate-180", 0, ROTATE_180_OPT, 0, 0 },
   { "rotate-270", 0, ROTATE_270_OPT, 0, 0 },
-  { "no-rotate", 0, NO_ROTATE_OPT, 0, Clp_LongMinMatch, 4 }, /****/
-  
+  { "no-rotate", 0, NO_ROTATE_OPT, 0, 0 },
+
+  { "scale", 0, SCALE_OPT, SCALE_FACTOR_TYPE, Clp_Negate },
   { "screen", 0, LOGICAL_SCREEN_OPT, DIMENSIONS_TYPE, Clp_Negate },
   { "same-background", 0, SAME_BACKGROUND_OPT, 0, 0 },
   { "same-bg", 0, SAME_BACKGROUND_OPT, 0, 0 },
@@ -282,7 +293,7 @@ set_mode(int newmode)
   else if (mode == newmode)
     ;
   else
-    error("too late to change modes");
+    fatal_error("too late to change modes");
 }
 
 
@@ -295,7 +306,7 @@ set_frame_change(int kind)
   if (mode == BLANK_MODE)
     set_mode(MERGING);
   if (mode < DELETING && frames_done) {
-    error("frame selection and frame changes don't mix");
+    fatal_error("frame selection and frame changes don't mix");
     return;
   }
   assert(!nested_mode);
@@ -406,12 +417,17 @@ input_stream(char *name)
   files_given++;
   
   if (name == 0 || strcmp(name, "-") == 0) {
+#if defined(_MSDOS) || defined(_WIN32)
+    _setmode(_fileno(stdin), _O_BINARY);
+#elif defined(__EMX__)
+    _fsetmode(stdin, "b");
+#endif
     f = stdin;
     name = "<stdin>";
   } else
     f = fopen(name, "rb");
   if (!f) {
-    error("%s: %s.", name, strerror(errno));
+    error("%s: %s", name, strerror(errno));
     return;
   }
   
@@ -438,7 +454,7 @@ input_stream(char *name)
   gfs->userflags = 97; /* to indicate no output done */
   
   if (gfs->errors)
-    warning("%s contained some errors", name);
+    error("%s is partially corrupted", name);
   
   /* Processing when we've got a new input frame */
   if (mode == BLANK_MODE)
@@ -452,7 +468,7 @@ input_stream(char *name)
     else if (mode == EXPLODING) {
       /* Explode into current directory. Use name instead of input_name: it's
          #stdin# if required. */
-      char *slash = strrchr(name, '/');
+      char *slash = strrchr(name, PATHNAME_SEPARATOR);
       if (slash)
 	active_output_data.output_name = slash + 1;
       else
@@ -644,20 +660,24 @@ static void
 write_stream(char *output_name, Gif_Stream *gfs)
 {
   FILE *f;
-#ifndef NO_TTY_CHECK
-  extern int isatty(int);
-#endif
   
   if (output_name)
     f = fopen(output_name, "wb");
   else {
-#ifndef NO_TTY_CHECK
+#ifndef OUTPUT_GIF_TO_TERMINAL
+    extern int isatty(int);
     if (isatty(fileno(stdout))) {
       error("not writing to <stdout>: it's a terminal");
       return;
     }
+#endif
+#if defined(_MSDOS) || defined(_WIN32)
+    _setmode(_fileno(stdout), _O_BINARY);
+#elif defined(__EMX__)
+    _fsetmode(stdout, "b");
 #endif
     f = stdout;
+    output_name = "<stdout>";
   }
   
   if (f) {
@@ -665,7 +685,7 @@ write_stream(char *output_name, Gif_Stream *gfs)
     fclose(f);
     any_output_successful = 1;
   } else
-    error("%s: %s.", output_name ? output_name : "<stdout>", strerror(errno));
+    error("%s: %s", output_name, strerror(errno));
 }
 
 static void
@@ -680,7 +700,7 @@ merge_and_write_frames(char *outfile, int f1, int f2)
   colormap_change = active_output_data.colormap_size > 0
     || active_output_data.colormap_fixed;
   compress_immediately = !colormap_change
-    && active_output_data.resize_width <= 0
+    && active_output_data.scaling == 0
     && active_output_data.optimizing <= 0;
   warn_local_colormaps = !colormap_change;
   
@@ -688,9 +708,12 @@ merge_and_write_frames(char *outfile, int f1, int f2)
 			     compress_immediately);
   
   if (out) {
-    if (active_output_data.resize_width > 0)
+    if (active_output_data.scaling == 1)
       resize_stream(out, active_output_data.resize_width,
 		    active_output_data.resize_height);
+    else if (active_output_data.scaling == 2)
+      resize_stream(out, active_output_data.scale_x * out->screen_width,
+		    active_output_data.scale_y * out->screen_height);
     if (colormap_change)
       do_colormap_change(out);
     if (output_transforms)
@@ -857,9 +880,7 @@ initialize_def_frame(void)
   def_output_data.colormap_dither = 0;
   
   def_output_data.optimizing = 0;
-  
-  def_output_data.resize_width = 0;
-  def_output_data.resize_height = 0;
+  def_output_data.scaling = 0;
   
   active_output_data = def_output_data;
 }
@@ -886,12 +907,8 @@ combine_output_options(void)
   COMBINE_ONE_OUTPUT_OPTION(CH_LOOPCOUNT, loopcount);
   
   COMBINE_ONE_OUTPUT_OPTION(CH_OPTIMIZE, optimizing);
-  
-  if (CHANGED(recent, CH_COLORMAP)) {
-    MARK_CH(output, CH_COLORMAP);
-    active_output_data.colormap_size = def_output_data.colormap_size;
-    active_output_data.colormap_algorithm = def_output_data.colormap_algorithm;
-  }
+  COMBINE_ONE_OUTPUT_OPTION(CH_COLORMAP, colormap_size);
+  COMBINE_ONE_OUTPUT_OPTION(CH_COLORMAP_METHOD, colormap_algorithm);
   if (CHANGED(recent, CH_USE_COLORMAP)) {
     MARK_CH(output, CH_USE_COLORMAP);
     if (def_output_data.colormap_fixed)
@@ -903,8 +920,11 @@ combine_output_options(void)
   
   if (CHANGED(recent, CH_RESIZE)) {
     MARK_CH(output, CH_RESIZE);
+    active_output_data.scaling = def_output_data.scaling;
     active_output_data.resize_width = def_output_data.resize_width;
     active_output_data.resize_height = def_output_data.resize_height;
+    active_output_data.scale_x = def_output_data.scale_x;
+    active_output_data.scale_y = def_output_data.scale_y;
   }
   
   def_output_data.colormap_fixed = 0;
@@ -972,11 +992,13 @@ main(int argc, char **argv)
      0);
   Clp_AddType(clp, DIMENSIONS_TYPE, 0, parse_dimensions, 0);
   Clp_AddType(clp, POSITION_TYPE, 0, parse_position, 0);
+  Clp_AddType(clp, SCALE_FACTOR_TYPE, 0, parse_scale_factor, 0);
   Clp_AddType(clp, FRAME_SPEC_TYPE, 0, parse_frame_spec, 0);
   Clp_AddType(clp, COLOR_TYPE, Clp_DisallowOptions, parse_color, 0);
   Clp_AddType(clp, RECTANGLE_TYPE, 0, parse_rectangle, 0);
   Clp_AddType(clp, TWO_COLORS_TYPE, Clp_DisallowOptions, parse_two_colors, 0);
   Clp_SetOptionChar(clp, '+', Clp_ShortNegated);
+  Clp_SetErrorHandler(clp, clp_error_handler);
   
   program_name = Clp_ProgramName(clp);
   
@@ -984,7 +1006,7 @@ main(int argc, char **argv)
   initialize_def_frame();
   
 #ifdef DMALLOC
-  dmalloc_verbose(fopen("fudge", "w"));
+  dmalloc_verbose("fudge");
 #endif
   
   /* Yep, I'm an idiot.
@@ -1287,7 +1309,7 @@ main(int argc, char **argv)
       if (clp->negated)
 	def_output_data.loopcount = -1;
       else
-	def_output_data.loopcount = clp->hadarg ? clp->val.i : 0;
+	def_output_data.loopcount = (clp->have_arg ? clp->val.i : 0);
       break;
       
      case SAME_LOOPCOUNT_OPT:
@@ -1299,10 +1321,8 @@ main(int argc, char **argv)
       MARK_CH(output, CH_OPTIMIZE);
       if (clp->negated)
 	def_output_data.optimizing = 0;
-      else if (clp->hadarg)
-	def_output_data.optimizing = clp->val.i;
       else
-	def_output_data.optimizing = 1;
+	def_output_data.optimizing = (clp->have_arg ? clp->val.i : 1);
       break;
       
      case UNOPTIMIZE_OPT:
@@ -1363,7 +1383,7 @@ main(int argc, char **argv)
       break;
       
      case COLORMAP_ALGORITHM_OPT:
-      MARK_CH(output, CH_COLORMAP);
+      MARK_CH(output, CH_COLORMAP_METHOD);
       def_output_data.colormap_algorithm = clp->val.i;
       break;
       
@@ -1375,30 +1395,45 @@ main(int argc, char **argv)
      case RESIZE_OPT:
       MARK_CH(output, CH_RESIZE);
       if (clp->negated)
-	def_output_data.resize_width = def_output_data.resize_height = 0;
+	def_output_data.scaling = 0;
       else if (dimensions_x <= 0 || dimensions_y <= 0) {
 	error("`--resize' width and height must be positive");
-	def_output_data.resize_width = def_output_data.resize_height = 0;
+	def_output_data.scaling = 0;
       } else {
+	def_output_data.scaling = 1; /* use resize dimensions */
 	def_output_data.resize_width = dimensions_x;
 	def_output_data.resize_height = dimensions_y;
       }
       break;
       
+     case SCALE_OPT:
+      MARK_CH(output, CH_RESIZE);
+      if (clp->negated)
+	def_output_data.scaling = 0;
+      else if (parsed_scale_factor_x <= 0 || parsed_scale_factor_y <= 0) {
+	error("`--scale' X and Y factors must be positive");
+	def_output_data.scaling = 0;
+      } else {
+	def_output_data.scaling = 2; /* use scale factor */
+	def_output_data.scale_x = parsed_scale_factor_x;
+	def_output_data.scale_y = parsed_scale_factor_y;
+      }
+      break;
+      
       /* RANDOM OPTIONS */
       
      case VERSION_OPT:
-      printf("Gifsicle version %s\n", VERSION);
-      printf("Copyright (C) 1997-8 Eddie Kohler\n\
+      printf("LCDF Gifsicle %s\n", VERSION);
+      printf("Copyright (C) 1997-9 Eddie Kohler\n\
 This is free software; see the source for copying conditions.\n\
 There is NO warranty, not even for merchantability or fitness for a\n\
 particular purpose.\n");
-      exit(0);
+      exit(EXIT_OK);
       break;
       
      case HELP_OPT:
       usage();
-      exit(0);
+      exit(EXIT_OK);
       break;
       
      case OUTPUT_OPT:
@@ -1426,7 +1461,7 @@ particular purpose.\n");
      bad_option:
      case Clp_BadOption:
       short_usage();
-      exit(1);
+      exit(EXIT_USER_ERR);
       break;
       
      default:
@@ -1456,5 +1491,5 @@ particular purpose.\n");
 #ifdef DMALLOC
   dmalloc_report();
 #endif
-  return (any_output_successful ? 0 : 1);
+  return (error_count ? EXIT_ERR : EXIT_OK);
 }
