@@ -1,5 +1,5 @@
 /* optimize.c - Functions to optimize animated GIFs.
-   Copyright (C) 1997-8 Eddie Kohler, eddietwo@lcs.mit.edu
+   Copyright (C) 1997-9 Eddie Kohler, eddietwo@lcs.mit.edu
    This file is part of gifsicle.
 
    Gifsicle is free software. It is distributed under the GNU Public License,
@@ -7,6 +7,7 @@
    as this notice is kept intact and this source code is made available. There
    is no warranty, express or implied. */
 
+#include <config.h>
 #include "gifsicle.h"
 #include <assert.h>
 #include <string.h>
@@ -38,7 +39,7 @@ static Gif_Colormap *in_global_map;
 static Gif_Colormap *out_global_map;
 
 #define TRANSP (0)
-static int background;
+static u_int16_t background;
 #define NOT_IN_OUT_GLOBAL (256)
 static u_int16_t *last_data;
 static u_int16_t *this_data;
@@ -131,8 +132,7 @@ permuting_sorter_down(const void *v1, const void *v2)
 }
 
 static u_int16_t *
-sort_permutation(u_int16_t *perm, u_int16_t size, int32_t *values,
-		 int is_down)
+sort_permutation(u_int16_t *perm, int size, int32_t *values, int is_down)
 {
   permuting_sort_values = values;
   if (is_down)
@@ -1287,7 +1287,7 @@ finalize_optimizer(Gif_Stream *gfs)
   int i;
   
   if (background == TRANSP)
-    gfs->background = gfs->images[0]->transparent;
+    gfs->background = (byte)gfs->images[0]->transparent;
 
   /* 10.Dec.1998 - prefer GIF_DISPOSAL_NONE to GIF_DISPOSAL_ASIS. This is
      semantically "wrong" -- it's better to set the disposal explicitly than
